#   Copyright (c) 2021, Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import secrets
import string
import os
import sys
from typing import Optional
import hype

def generate_password(
    ascii_letters: string.ascii_letters = None,
    digits: string.digits = None,
    length: int = 16
):
    """
    Generate a random and cryptographically safe password
    based on `secrets` module.

    """

    if not ascii_letters:
        ascii_letters = string.ascii_letters
    if not digits:
        digits = string.digits

    chars = ascii_letters + digits

    return ''.join(
        secrets.choice(chars)
        for i in range(length)
    )

def create_hidden_folder(
    folder: str,
    attribute: int = 0x02
):
    """
    Create a hidden folder for Win32.
    On Unix machines, we just need to add '.'
    infront of the folder.
    """
    
    folder = str(folder)
    FILE_ATTRIBUTE_HIDDEN = attribute
    
    if sys.platform == 'linux' or sys.platform == 'darwin':
        if not folder.startswith('.'):
            os.mkdir('.%s' % (folder))
        else:
            os.mkdir(folder)

    elif sys.platform == 'win32':
        import ctypes

        os.mkdir(folder)
        ret = ctypes.windll.kernel32.SetFileAttributesW(
            folder,
            FILE_ATTRIBUTE_HIDDEN
        )

        if not ret:
            raise ctypes.WinError()

def hide_file(
    file: str,
    attribute: int = 0x02
):
    """
    Create a hidden file for Win32.
    On Unix machines, we just need to add '.'
    infront of the file.
    """
    
    file = str(file)

    if attribute == 0x80:
        HIDE = False
        FILE_ATTRIBUTE = attribute
    else:
        HIDE = True
        FILE_ATTRIBUTE = attribute
    
    if sys.platform == 'linux' or sys.platform == 'darwin':
        if not file.startswith('.') and not HIDE:
            os.rename('.%s' % (file))
        else:
            os.rename(file)

    elif sys.platform == 'win32':
        import ctypes

        ret = ctypes.windll.kernel32.SetFileAttributesW(
            file,
            FILE_ATTRIBUTE
        )

        if not ret:
            raise ctypes.WinError()
    
def print_banner(color: Optional[str] = 'magenta'):
    hype.print("""\n[%s] o8o    .o   .ooooo.                   oooo                                   \n `"'  o888  d88'   `8.                 `888                                   \noooo   888  Y88..  .8' ooo. .oo.        888   .oooo.   ooo. .oo.    .oooooooo \n`888   888   `88888b.  `888P"Y88b       888  `P  )88b  `888P"Y88b  888' `88b  \n 888   888  .8'  ``88b  888   888       888   .oP"888   888   888  888   888  \n 888   888  `8.   .88P  888   888  .o.  888  d8(  888   888   888  `88bod8P'  \no888o o888o  `boood8'  o888o o888o Y8P o888o `Y888""8o o888o o888o `8oooooo.  \n                                                                   d"     YD  \n                                                                   "Y88888P'\n                [blue]- Build app internationally![/blue]\n                [green]- Made with [red]<3[/red][green] by Zenqi (https://github.com/znqi)[/green]\n\n[/%s]""" % (color, color))

def clear_screen():
    
    if sys.platform == 'win32':
        os.system('cls')
    elif sys.platform == 'linux' or sys.platform == 'linux2':
        os.system('clear')
    elif sys.platform == 'darwin':
        os.system('clear')



#   Copyright (c) 2021, Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

try:
  import googletrans
  TRANSLATOR = googletrans.Translator()
  
except ModuleNotFoundError:
  TRANSLATOR = None

import pathlib

BASE_DIR: pathlib.Path = pathlib.Path().resolve()
COUNTRIES_TO_LANG = {
    
  "afrikaans": "af",
  "afrikaans (south africa)": "af_ZA",
  "arabic": "ar",
  "arabic (u.a.e.)": "ar_AE",
  "arabic (bahrain)": "ar_BH",
  "arabic (algeria)": "ar_DZ",
  "arabic (egypt)": "ar_EG",
  "arabic (iraq)": "ar_IQ",
  "arabic (jordan)": "ar_JO",
  "arabic (kuwait)": "ar_KW",
  "arabic (lebanon)": "ar_LB",
  "arabic (libya)": "ar_LY",
  "arabic (morocco)": "ar_MA",
  "arabic (oman)": "ar_OM",
  "arabic (qatar)": "ar_QA",
  "arabic (saudi arabia)": "ar_SA",
  "arabic (syria)": "ar_SY",
  "arabic (tunisia)": "ar_TN",
  "arabic (yemen)": "ar_YE",
  "azeri (latin)": "az",
  "azeri (latin) (azerbaijan)": "az_AZ",
  "azeri (cyrillic) (azerbaijan)": "az_AZ",
  "belarusian": "be",
  "belarusian (belarus)": "be_BY",
  "bulgarian": "bg",
  "bulgarian (bulgaria)": "bg_BG",
  "bosnian (bosnia and herzegovina)": "bs_BA",
  "catalan": "ca",
  "catalan (spain)": "ca_ES",
  "czech": "cs",
  "czech (czech republic)": "cs_CZ",
  "welsh": "cy",
  "welsh (united kingdom)": "cy_GB",
  "danish": "da",
  "danish (denmark)": "da_DK",
  "german": "de",
  "german (austria)": "de_AT",
  "german (switzerland)": "de_CH",
  "german (germany)": "de_DE",
  "german (liechtenstein)": "de_LI",
  "german (luxembourg)": "de_LU",
  "divehi": "dv",
  "divehi (maldives)": "dv_MV",
  "greek": "el",
  "greek (greece)": "el_GR",
  "english": "en",
  "english (australia)": "en_AU",
  "english (belize)": "en_BZ",
  "english (canada)": "en_CA",
  "english (caribbean)": "en_CB",
  "english (united kingdom)": "en_GB",
  "english (ireland)": "en_IE",
  "english (jamaica)": "en_JM",
  "english (new zealand)": "en_NZ",
  "english (republic of the philippines)": "en_PH",
  "english (trinidad and tobago)": "en_TT",
  "english (united states)": "en_US",
  "english (south africa)": "en_ZA",
  "english (zimbabwe)": "en_ZW",
  "esperanto": "eo",
  "spanish": "es",
  "spanish (argentina)": "es_AR",
  "spanish (bolivia)": "es_BO",
  "spanish (chile)": "es_CL",
  "spanish (colombia)": "es_CO",
  "spanish (costa rica)": "es_CR",
  "spanish (dominican republic)": "es_DO",
  "spanish (ecuador)": "es_EC",
  "spanish (castilian)": "es_ES",
  "spanish (spain)": "es_ES",
  "spanish (guatemala)": "es_GT",
  "spanish (honduras)": "es_HN",
  "spanish (mexico)": "es_MX",
  "spanish (nicaragua)": "es_NI",
  "spanish (panama)": "es_PA",
  "spanish (peru)": "es_PE",
  "spanish (puerto rico)": "es_PR",
  "spanish (paraguay)": "es_PY",
  "spanish (el salvador)": "es_SV",
  "spanish (uruguay)": "es_UY",
  "spanish (venezuela)": "es_VE",
  "estonian": "et",
  "estonian (estonia)": "et_EE",
  "basque": "eu",
  "basque (spain)": "eu_ES",
  "farsi": "fa",
  "farsi (iran)": "fa_IR",
  "finnish": "fi",
  "finnish (finland)": "fi_FI",
  "faroese": "fo",
  "faroese (faroe islands)": "fo_FO",
  "french": "fr",
  "french (belgium)": "fr_BE",
  "french (canada)": "fr_CA",
  "french (switzerland)": "fr_CH",
  "french (france)": "fr_FR",
  "french (luxembourg)": "fr_LU",
  "french (principality of monaco)": "fr_MC",
  "galician": "gl",
  "galician (spain)": "gl_ES",
  "gujarati": "gu",
  "gujarati (india)": "gu_IN",
  "hebrew": "he",
  "hebrew (israel)": "he_IL",
  "hindi": "hi",
  "hindi (india)": "hi_IN",
  "croatian": "hr",
  "croatian (bosnia and herzegovina)": "hr_BA",
  "croatian (croatia)": "hr_HR",
  "hungarian": "hu",
  "hungarian (hungary)": "hu_HU",
  "armenian": "hy",
  "armenian (armenia)": "hy_AM",
  "indonesian": "id",
  "indonesian (indonesia)": "id_ID",
  "icelandic": "is",
  "icelandic (iceland)": "is_IS",
  "italian": "it",
  "italian (switzerland)": "it_CH",
  "italian (italy)": "it_IT",
  "japanese": "ja",
  "japanese (japan)": "ja_JP",
  "georgian": "ka",
  "georgian (georgia)": "ka_GE",
  "kazakh": "kk",
  "kazakh (kazakhstan)": "kk_KZ",
  "kannada": "kn",
  "kannada (india)": "kn_IN",
  "korean": "ko",
  "korean (korea)": "ko_KR",
  "konkani": "kok",
  "konkani (india)": "kok_IN",
  "kyrgyz": "ky",
  "kyrgyz (kyrgyzstan)": "ky_KG",
  "lithuanian": "lt",
  "lithuanian (lithuania)": "lt_LT",
  "latvian": "lv",
  "latvian (latvia)": "lv_LV",
  "maori": "mi",
  "maori (new zealand)": "mi_NZ",
  "fyro macedonian": "mk",
  "fyro macedonian (former yugoslav republic of macedonia)": "mk_MK",
  "mongolian": "mn",
  "mongolian (mongolia)": "mn_MN",
  "marathi": "mr",
  "marathi (india)": "mr_IN",
  "malay": "ms",
  "malay (brunei darussalam)": "ms_BN",
  "malay (malaysia)": "ms_MY",
  "maltese": "mt",
  "maltese (malta)": "mt_MT",
  "norwegian (bokm?l)": "nb",
  "norwegian (bokm?l) (norway)": "nb_NO",
  "dutch": "nl",
  "dutch (belgium)": "nl_BE",
  "dutch (netherlands)": "nl_NL",
  "norwegian (nynorsk) (norway)": "nn_NO",
  "northern sotho": "ns",
  "northern sotho (south africa)": "ns_ZA",
  "punjabi": "pa",
  "punjabi (india)": "pa_IN",
  "polish": "pl",
  "polish (poland)": "pl_PL",
  "pashto": "ps",
  "pashto (afghanistan)": "ps_AR",
  "portuguese": "pt",
  "portuguese (brazil)": "pt_BR",
  "portuguese (portugal)": "pt_PT",
  "quechua": "qu",
  "quechua (bolivia)": "qu_BO",
  "quechua (ecuador)": "qu_EC",
  "quechua (peru)": "qu_PE",
  "romanian": "ro",
  "romanian (romania)": "ro_RO",
  "russian": "ru",
  "russian (russia)": "ru_RU",
  "sanskrit": "sa",
  "sanskrit (india)": "sa_IN",
  "sami (northern)": "se",
  "sami (northern) (finland)": "se_FI",
  "sami (skolt) (finland)": "se_FI",
  "sami (inari) (finland)": "se_FI",
  "sami (northern) (norway)": "se_NO",
  "sami (lule) (norway)": "se_NO",
  "sami (southern) (norway)": "se_NO",
  "sami (northern) (sweden)": "se_SE",
  "sami (lule) (sweden)": "se_SE",
  "sami (southern) (sweden)": "se_SE",
  "slovak": "sk",
  "slovak (slovakia)": "sk_SK",
  "slovenian": "sl",
  "slovenian (slovenia)": "sl_SI",
  "albanian": "sq",
  "albanian (albania)": "sq_AL",
  "serbian (latin) (bosnia and herzegovina)": "sr_BA",
  "serbian (cyrillic) (bosnia and herzegovina)": "sr_BA",
  "serbian (latin) (serbia and montenegro)": "sr_SP",
  "serbian (cyrillic) (serbia and montenegro)": "sr_SP",
  "swedish": "sv",
  "swedish (finland)": "sv_FI",
  "swedish (sweden)": "sv_SE",
  "swahili": "sw",
  "swahili (kenya)": "sw_KE",
  "syriac": "syr",
  "syriac (syria)": "syr_SY",
  "tamil": "ta",
  "tamil (india)": "ta_IN",
  "telugu": "te",
  "telugu (india)": "te_IN",
  "thai": "th",
  "thai (thailand)": "th_TH",
  "tagalog": "tl",
  "tagalog (philippines)": "tl_PH",
  "tswana": "tn",
  "tswana (south africa)": "tn_ZA",
  "turkish": "tr",
  "turkish (turkey)": "tr_TR",
  "tatar": "tt",
  "tatar (russia)": "tt_RU",
  "tsonga": "ts",
  "ukrainian": "uk",
  "ukrainian (ukraine)": "uk_UA",
  "urdu": "ur",
  "urdu (islamic republic of pakistan)": "ur_PK",
  "uzbek (latin)": "uz",
  "uzbek (latin) (uzbekistan)": "uz_UZ",
  "uzbek (cyrillic) (uzbekistan)": "uz_UZ",
  "vietnamese": "vi",
  "vietnamese (viet nam)": "vi_VN",
  "xhosa": "xh",
  "xhosa (south africa)": "xh_ZA",
  "chinese": "zh_CN",
  "chinese (traditional)": "zh_TW",
  "zulu": "zu",
  "zulu (south africa)": "zu_ZA"
}

COUNTRIES_TO_CODE = {
    "aruba": "aw",
    "afghanistan": "af",
    "angola": "ao",
    "anguilla": "ai",
    "\u00e5land islands": "ax",
    "albania": "al",
    "andorra": "ad",
    "united arab emirates": "ae",
    "argentina": "ar",
    "armenia": "am",
    "american samoa": "as",
    "antarctica": "aq",
    "french southern territories": "tf",
    "antigua and barbuda": "ag",
    "australia": "au",
    "austria": "at",
    "azerbaijan": "az",
    "burundi": "bi",
    "belgium": "be",
    "benin": "bj",
    "bonaire, sint eustatius and saba": "bq",
    "burkina faso": "bf",
    "bangladesh": "bd",
    "bulgaria": "bg",
    "bahrain": "bh",
    "bahamas": "bs",
    "bosnia and herzegovina": "ba",
    "saint barth\u00e9lemy": "bl",
    "belarus": "by",
    "belize": "bz",
    "bermuda": "bm",
    "bolivia, plurinational state of": "bo",
    "brazil": "br",
    "barbados": "bb",
    "brunei darussalam": "bn",
    "bhutan": "bt",
    "bouvet island": "bv",
    "botswana": "bw",
    "central african republic": "cf",
    "canada": "ca",
    "cocos (keeling) islands": "cc",
    "switzerland": "ch",
    "chile": "cl",
    "china": "cn",
    "c\u00f4te d'ivoire": "ci",
    "cameroon": "cm",
    "congo, the democratic republic of the": "cd",
    "congo": "cg",
    "cook islands": "ck",
    "colombia": "co",
    "comoros": "km",
    "cabo verde": "cv",
    "costa rica": "cr",
    "cuba": "cu",
    "cura\u00e7ao": "cw",
    "christmas island": "cx",
    "cayman islands": "ky",
    "cyprus": "cy",
    "czechia": "cz",
    "germany": "de",
    "djibouti": "dj",
    "dominica": "dm",
    "denmark": "dk",
    "dominican republic": "do",
    "algeria": "dz",
    "ecuador": "ec",
    "egypt": "eg",
    "eritrea": "er",
    "western sahara": "eh",
    "spain": "es",
    "estonia": "ee",
    "ethiopia": "et",
    "finland": "fi",
    "fiji": "fj",
    "falkland islands (malvinas)": "fk",
    "france": "fr",
    "faroe islands": "fo",
    "micronesia, federated states of": "fm",
    "gabon": "ga",
    "united kingdom": "gb",
    "georgia": "ge",
    "guernsey": "gg",
    "ghana": "gh",
    "gibraltar": "gi",
    "guinea": "gn",
    "guadeloupe": "gp",
    "gambia": "gm",
    "guinea_bissau": "gw",
    "equatorial guinea": "gq",
    "greece": "gr",
    "grenada": "gd",
    "greenland": "gl",
    "guatemala": "gt",
    "french guiana": "gf",
    "guam": "gu",
    "guyana": "gy",
    "hong kong": "hk",
    "heard island and mcdonald islands": "hm",
    "honduras": "hn",
    "croatia": "hr",
    "haiti": "ht",
    "hungary": "hu",
    "indonesia": "id",
    "isle of man": "im",
    "india": "in",
    "british indian ocean territory": "io",
    "ireland": "ie",
    "iran, islamic republic of": "ir",
    "iraq": "iq",
    "iceland": "is",
    "israel": "il",
    "italy": "it",
    "jamaica": "jm",
    "jersey": "je",
    "jordan": "jo",
    "japan": "jp",
    "kazakhstan": "kz",
    "kenya": "ke",
    "kyrgyzstan": "kg",
    "cambodia": "kh",
    "kiribati": "ki",
    "saint kitts and nevis": "kn",
    "korea, republic of": "kr",
    "kuwait": "kw",
    "lao people's democratic republic": "la",
    "lebanon": "lb",
    "liberia": "lr",
    "libya": "ly",
    "saint lucia": "lc",
    "liechtenstein": "li",
    "sri lanka": "lk",
    "lesotho": "ls",
    "lithuania": "lt",
    "luxembourg": "lu",
    "latvia": "lv",
    "macao": "mo",
    "saint martin (french part)": "mf",
    "morocco": "ma",
    "monaco": "mc",
    "moldova, republic of": "md",
    "madagascar": "mg",
    "maldives": "mv",
    "mexico": "mx",
    "marshall islands": "mh",
    "north macedonia": "mk",
    "mali": "ml",
    "malta": "mt",
    "myanmar": "mm",
    "montenegro": "me",
    "mongolia": "mn",
    "northern mariana islands": "mp",
    "mozambique": "mz",
    "mauritania": "mr",
    "montserrat": "ms",
    "martinique": "mq",
    "mauritius": "mu",
    "malawi": "mw",
    "malaysia": "my",
    "mayotte": "yt",
    "namibia": "na",
    "new caledonia": "nc",
    "niger": "ne",
    "norfolk island": "nf",
    "nigeria": "ng",
    "nicaragua": "ni",
    "niue": "nu",
    "netherlands": "nl",
    "norway": "no",
    "nepal": "np",
    "nauru": "nr",
    "new zealand": "nz",
    "oman": "om",
    "pakistan": "pk",
    "panama": "pa",
    "pitcairn": "pn",
    "peru": "pe",
    "philippines": "ph",
    "palau": "pw",
    "papua new guinea": "pg",
    "poland": "pl",
    "puerto rico": "pr",
    "korea, democratic people's republic of": "kp",
    "portugal": "pt",
    "paraguay": "py",
    "palestine, state of": "ps",
    "french polynesia": "pf",
    "qatar": "qa",
    "r\u00e9union": "re",
    "romania": "ro",
    "russian federation": "ru",
    "rwanda": "rw",
    "saudi arabia": "sa",
    "sudan": "sd",
    "senegal": "sn",
    "singapore": "sg",
    "south georgia and the south sandwich islands": "gs",
    "saint helena, ascension and tristan da cunha": "sh",
    "svalbard and jan mayen": "sj",
    "solomon islands": "sb",
    "sierra leone": "sl",
    "el salvador": "sv",
    "san marino": "sm",
    "somalia": "so",
    "saint pierre and miquelon": "pm",
    "serbia": "rs",
    "south sudan": "ss",
    "sao tome and principe": "st",
    "suriname": "sr",
    "slovakia": "sk",
    "slovenia": "si",
    "sweden": "se",
    "eswatini": "sz",
    "sint maarten (dutch part)": "sx",
    "seychelles": "sc",
    "syrian arab republic": "sy",
    "turks and caicos islands": "tc",
    "chad": "td",
    "togo": "tg",
    "thailand": "th",
    "tajikistan": "tj",
    "tokelau": "tk",
    "turkmenistan": "tm",
    "timor_leste": "tl",
    "tonga": "to",
    "trinidad and tobago": "tt",
    "tunisia": "tn",
    "turkey": "tr",
    "tuvalu": "tv",
    "taiwan, province of china": "tw",
    "tanzania, united republic of": "tz",
    "uganda": "ug",
    "ukraine": "ua",
    "united states minor outlying islands": "um",
    "uruguay": "uy",
    "united states": "us",
    "uzbekistan": "uz",
    "holy see (vatican city state)": "va",
    "saint vincent and the grenadines": "vc",
    "venezuela, bolivarian republic of": "ve",
    "virgin islands, british": "vg",
    "virgin islands, u.s.": "vi",
    "viet nam": "vn",
    "vanuatu": "vu",
    "wallis and futuna": "wf",
    "samoa": "ws",
    "yemen": "ye",
    "south africa": "za",
    "zambia": "zm",
    "zimbabwe": "zw"
}



#   Copyright (c) 2021, Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import Optional


class i18nError(Exception):
    pass

class LanguageError(i18nError):
    pass

class ConfigError(i18nError):
    pass

class MetaDataError(i18nError):
    pass

class PluginError(i18nError):
    pass


def throw(
    error_cls: i18nError, 
    message: str, 
    code: str, 
    line: Optional[str] = '─',
):

    """
    A error throwing with format given the error class,
    message and the code.
    """

    raise error_cls("\n\n%s\n\t[ERROR: %s] %s\n%s\n\n"
    %(
        line*70,
        code,
        message,
        line*70
    ))


ERROR_CODES = {
    'unknown_error': '0000',
    'not_found': '0001',
    'installation_error': '0002',
    'internal_error': '0003',
    'external_error': '0004',
    'package_error': '0005',
    'deprecation_error': '0006' 
}



#   Copyright (c) 2021, Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#: Convert https://github.com/znqi/sidle into a single python file
#: With some changes

from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto import Random

from typing import (
    Any,
    List,
    Optional,
    Iterable,
    AnyStr,
    TextIO,
    Tuple
)
import ast
import os


class Sidle:
    """
    Sidle is a secure data storing with password and 
    simple encryption. This class handles all sidle 
    operation that can be storing sidle data on a 
    filename and more..
    Parameters:
        filename (str):
            The filename to be used for storing data
        password (Any):
            The password for the file in order to do 
            Sidle operation
        **options (Keyword Arguments):
            Additional option for the operation.
    """

    def __init__(
        self,
        filename: str,
        password: Any,
        **options
    ):

        #: The filename to be used for storing data
        self.filename = filename

        #: The password for the file in order to do 
        #: Sidle operation
        self.password = password
        
        #: A sidle data instance for handling data storing.
        self.sidle: SidleData = SidleData(
            password=self.password
        )

        #: Additional option for the operation.
        self.options = options

        self.__check_for_filename()

    def __getitem__(self, key: Any):
        """
        Get the item from the filename given.
        """

        raw = self.read_raw(self.filename, True)
        
        if len(raw) == 0:
            return None

        data = self.sidle.load(raw)
        return data.__getitem__(
            key=key
        )

    def __setitem__(self, key: Any, value: Any):
        """
        Set the item from the filename given
        """

        self.insert(
            key=key,
            value=value
        )


    def __check_for_filename(self):
        """
        Check if the given filename is already existed.
        If not then create a new one.
        """


        if not os.path.isfile(self.filename):
            with open(self.filename, 'wb') as f:
                pass
    
    def read_raw(
        self, 
        filename: str, 
        writable: Optional[bool] = False
    ) -> TextIO:
        
        """
        Read the raw file and return a TextIO object
        """

        if writable:
            mode = 'rb+'
        else:
            mode = 'rb'

        with open(filename, mode) as f:
            return f.read()


    def write_raw(self, filename: str, data: bytes) -> TextIO:
        """
        Write a raw data to the file
        """

        with open(filename, 'wb') as f:
            f.write(data)

    
    def read(self):
        """
        Return the representation of the
        SidleData.
        """

        return repr(self.sidle)

    def insert(
        self, 
        key: str, 
        value: Any
    ):
        """
        Insert the key and a value to the file.    
        """

        raw = self.read_raw(
            filename=self.filename
        )

        if len(raw) == 0:

            self.sidle.set(
                key=key,
                value=value
            )

            self.sidle.save(self.filename)

        else:
            data = self.sidle.load(raw)
            data.set(
                key=key,
                value=value
            )
                
            data.save(
                filename=self.filename
            )


    def __repr__(self):
        return self._get_repr_value()

    def _get_repr_value(self):
        raw = self.read_raw(
            filename=self.filename
        )
        

        if len(raw) == 0:
            _data = None
        else:
            _data = self.sidle.load(raw)
  
        
        name = type(self).__name__
        filename = self.filename
        password = password_with_asterisk(
            password=self.password
        )
        
        if _data:
            _len = len(_data)
            keys = [
                self.sidle._enc.decrypt(k)
                for k in _data.keys()
            ]
            

        else:
            _len = 0
            keys = None

        return "%s(filename=%s, password=%s, length=%s, keys=%s)" % (
            name,
            filename,
            password,
            _len,
            keys

        )

class SidleEncryption:
    """
    Simple Encrypting and Decrypting strings
    with password
    
    Parameter:
        password (Any):
            The password that can be used for both
            encryption and decryption
    """

    def __init__(self, password: Any):
        """
        Initialize SidleEncryption
        """
        self.password = convert_bytes(password)

    def encrypt(self, string: str):
        """
        Encrypt the given string and return
        encrypted bytes that can be decrypted with the
        given password.
        """

        string = convert_bytes(string)
        
        key = self.__password_to_key(self.password)
        IV = self.make_initialization_vector()
        encryptor = AES.new(key, AES.MODE_CBC, IV)

        # store the IV at the beginning and encrypt
        return IV + encryptor.encrypt(self.pad_string(string))

    def decrypt(self, encrypted_string: bytes):
        """
        Decrypt the given encrypted string
        """

        key = self.__password_to_key(self.password)
        
        # extract the IV from the beginning
        IV = encrypted_string[:AES.block_size]  
        decryptor = AES.new(key, AES.MODE_CBC, IV)

        _str = decryptor.decrypt(encrypted_string[AES.block_size:])
        
        try:
            string = convert_string(
                self.unpad_string(_str)
            )
        except UnicodeDecodeError:
            raise PasswordError('Password: %s is not a valid password' % (self.password))


        return string


    def make_initialization_vector(self):
        """
        An initialization vector (IV) is a fixed-size input to a cryptographic
        primitive that is typically required to be random or pseudorandom.
        Randomization is crucial for encryption schemes to achieve semantic 
        security, a property whereby repeated usage of the scheme under the 
        same key does not allow an attacker to infer relationships 
        between segments of the encrypted message.
        """
        return Random.new().read(AES.block_size)
    
    
    def unpad_string(self, string: str):
        """
        Unpad the given string
        """
        
        to_pad = string[0]
        if to_pad == 0:
            return string

        return string[1:-to_pad]

    def pad_string(self, string, chunk_size=AES.block_size):
        """
        Pad string the peculirarity that uses the first byte
        is used to store how much padding is applied
        """
        assert chunk_size  <= 256, 'We are using one byte to represent padding'
        to_pad = (chunk_size - (len(string) + 1)) % chunk_size
        return bytes([to_pad]) + string + bytes([0] * to_pad)

    def __password_to_key(self, password: Any):
        """
        Use SHA-256 over the password to get a proper-sized AES key.
        This hashes the password into a 256 bit string. 
        """

        return SHA256.new(password).digest()

class SidleData:
    """
    A data structure for handling Sidle data
    with encryption: `SidleEncryption.`
    Parameters:
        password (Any):
            The password that can be used for encryption.
    """

    _enc: SidleEncryption
    _list: list

    def __init__(
        self, 
        password: Any, 
        data: List[Tuple[Any]] = None
    ):
        
        self.password = password
        
        if data:
            self._list = data
        else:
            self._list = []

        self._enc = SidleEncryption(
            password=self.password
        )
        
    @property
    def raw(self) -> bytes:
        """
        Return the encrypted raw string of the data.
        """
        return self._enc.encrypt(str(self._list))

    def __getitem__(self, key: str, error: Optional[bool]=False):
        """
        Get the value of the key given. 
        Parameter:
            key (str):
                The key for the value to get
            error (Optional[bool]):
                Set if the function throws `KeyError`
        """
        encrypted_data = [
            (self._enc.decrypt(k), self._enc.decrypt(v))
            for k,v in self._list 
        ]

        if not error:
            if isinstance(key, int):
                return encrypted_data[key]
            
            elif isinstance(key, slice):
                return self.__class__(encrypted_data[key])
            
        _: str = key.lower()
        for k, v in encrypted_data:
            if k.lower() == _:
                return v

        #: If the error is true, raise `KeyError`
        #: exception if the key doesn't exitst
        
        if error:
            raise KeyError(key)

    def __setitem__(self, key: str, value: Any):
        """
        Set the value of the item. This magic method
        can be used on:
        
        >>> sidle = SidleData('password')
        >>> sidle['username'] = 'zenqi'
        """

        if isinstance(key, (slice, int)):
            if isinstance(key, int):
                value = [value]

            value = [
                (k, v)
                for k,v in value
            ]

            if isinstance(key, int):
                self._list[key] = value[0]
            else:
                self._list[key] = value

        else:
            self.set(
                key=key, 
                value=value
            )

    def __delitem__(self, key: str, index: Optional[bool] = True):
        """
        Delete the key and its value from the list
        of header.
        Parameter:
            key (str):
                The key to be deleted
            index (Optional[bool]):
                Set if the slicing can be index.
        """

        encrypted_data = [
            (self._enc.decrypt(k), self._enc.decrypt(v))
            for k,v in self._list 
        ]

        if index and isinstance(key, (int, slice)):
            del encrypted_data[key]
            return 

        _: str = key.lower()
        _new: list = []
        for k, v in encrypted_data:
            if k.lower() != _:
                _new.append((k,v))

        encrypted_data[:] = _new

        self._list = encrypted_data

    def __eq__(self, other):
        def lowered(item):
            return (item[0].lower(),) + item[1:]

        return other.__class__ is self.__class__ and set(
            map(lowered, other.__list)
        ) == set(map(lowered, self._list))

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __copy__(self):
        return self.copy()
    
    def __contains__(self, key):
        """
        Check if a key is present.
        
        Example:
            
        """
        
        try:
            self.__getitem__(key)
        except KeyError:
            return False
        return True

    def get(
        self, 
        key: str, 
        default: Optional[Any] = None, 
        type: Optional[Any] = None, 
        bytes: Optional[bool]=False
    ):
        """
        Get the given key and return. If the type is defined
        then return the given type.
        Parameter:
            key (str):
                The key for the value to get.
            default (Optional[any]):
                Set the default value if the key doesn't exist
            
            type (Optional[any]):
                Set the type of the given value.
            bytes (Optional[bool]):
                Set if the value will be converted as bytes or no
        
        Example:
            
        
        """
        try:
            _value = self.__getitem__(key)
        except KeyError:
            return default

        if bytes:
            _value = convert_bytes(_value)

        if type != None:
            try:
                return type(_value)
            except ValueError:
                return default

        return _value

    def set(
        self,
        key: str,
        value: Any
    ):

        _key = self._enc.encrypt(key)
        _value = self._enc.encrypt(value)

        if self._list == None:
            self._list.append(
                (_key, _value)
            )

            return

        _iter = iter(self._list)
        _k = key.lower()

        for i, (o_key, o_value) in enumerate(_iter):
            if self._enc.decrypt(o_key).lower() == _k:                
                self._list[i] = (_key, _value)
                break

        else:
            self._list.append((_key, _value))
            return

        self._list[i + 1:] = [x for x in _iter if x[0].lower() != _k]


    def add(self, 
        key: AnyStr, 
        value: Any
    ):
        """
        Add a new tuple of header to the list of headers.
        Parameters:
            key (str):
                The key to be added
            value (Any):
                The value for the key to be added
            **kwargs(Dict[str, Any]):
                A keyword arguments for additional 
                option for the headers.
        """

        self._list.append((key, value))

    def remove(self, key: str):
        """
        Remove the key and its value from the list
        of header.
        Parameter:
            key (str):
                The key to be removed.
        """

        return self.__delitem__(key, False)

    def copy(self):
        return self.__class__(self._list)

    def clear(self):
        """
        Clear all data from the list.
        """

        del self._list[:]

    def pop(self, key: Optional[str] = None, default: Optional[Any] = None):
        """
        Remove the key and return the index of the key.
        Parameter:
            key (Optional[str]):
                The key to be popped. 
            default (Optional[None]):
                The default value if the key is `None`
        """

        if not key:
            self._list.pop()
        
        if isinstance(key, int):
            return self._list.pop(key)
        
        try:
            _value = self[key]
            self.remove(key)
        except KeyError:
            if default == None:
                return default
            else:
                default(_value)
        
        return _value

    def setlist(self, key: str, values: Iterable):
        """
        Remove any existing values for the header and add new one.
        Similar  to `set`
        
        Parameters:
            key (str):
                The key for the header.
            values(Iterable):
                Any value that can be iterate. For ex: list, tuple, etc
        """

        if values:
            _iter = iter(values)
            self.set(
                key=key, 
                value=next(_iter)
            )

            for value in _iter:
                self.add(
                    key=key,
                    value=value
                )
        
        else:
            self.remove(
                key=key
            )

    def keys(self):
        """
        Return all keys from the list of data.
        """
        _keys = []
        for k, _ in self:
            _keys.append(k)

        return _keys

    def load(self, data: bytes):
        """
        Load the the given data and return 
        A SileData class for the data given.
        """
        
        _ = self._enc.decrypt(data)
        if not _:
            try:
                self.load(data)
            except RecursionError:
                pass
        
        if isinstance(_, bytes):
            raise PasswordError('Password: %s is not a valid password' % (self.password))


        raw = ast.literal_eval(_)
        self.clear()
        self._list = raw

        return self

    def save(
        self, 
        filename: str, 
        data: Optional[bytes] = None
    ):

        """
        Save the sidle data on the given file io and
        closes it. Please take note that the File IO
        should be a writable bytes instead of string.
        This encrypt the data into bytes and store 
        in a file.
        """

        if data:
            data = data
        else:
            data = self.raw

        with open(filename, 'wb') as f:
            f.write(data)

    def __repr__(self):
        return self._get_repr_value(
            name=type(self).__name__
        )

    def _get_repr_value(self, name: str):
        length = len(self)
        keys = [self._enc.decrypt(x) for x in self.keys()]
        pwd = password_with_asterisk(
            password=self.password
        )

        return "%s(password=%s, length=%s, keys=%s)" % (
            name,
            pwd,
            length,
            keys
        )


def convert_string(
    value: Any, 
    encoding: str='utf-8'
) -> str:

    """
    Convert the given value of bytes to string.
    """

    if isinstance(value, bytes):
        value = value.decode(
            encoding=encoding
        )

    else:
        value = str(value)

    return value

def convert_bytes(
    value: Any, 
    encoding: str='utf-8'
) -> bytes:

    """
    Convert the given value of bytes to string.
    """
    if isinstance(value, str):
        value = value.encode(
            encoding=encoding
        )

    else:
        value = bytes(
            value,
            encoding=encoding
        )

    return value

def password_with_asterisk(
    password: str, 
    percentage: Optional[float] = 40
):
    
    if len(password) <= 2:
        return "*" * len(password)
  
    percent = int(
        (percentage * len(password)) / 100.0
    )
        
    _ = "*" * (len(password) - percent)
    return "%s%s" % (_, password[-percent:])

class SidleError(Exception):
    pass

class PasswordError(SidleError):
    pass




#   Copyright (c) 2021, Zenqi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import os
import time
import shutil
import pathlib
import locale
import configparser

from ._sidle import Sidle
from .errors import (
    ConfigError,
    LanguageError,
    ERROR_CODES,
    MetaDataError,
    PluginError,
    throw
)
from typing import (
    Any,
    Optional,
    Sequence
)

from .constants import (
    BASE_DIR,
    COUNTRIES_TO_LANG,
    TRANSLATOR
)
from .utils import (
    clear_screen,
    generate_password,
    create_hidden_folder,
    hide_file,
    print_banner
)
from .logs import logger

import hype.ui as ui
import hype

class i18nLang:
    """
    
    `i18nLang` is a wrapper for getting value from
    the compiled, build output of a `i18nLangBuilder`
    This class handles all build data using `Sidle`.

    Parameter:
    ---
        lang_folder (str):
            The path for the compiled lang folder.

        lang (Optional[str]):
            The default lang to be used

        key (Optional[str]):
            The key for the compiled lang pack.
            This is available @ .build.ini file

    Example:
    ---
        >>> from i18n.lang import i18nLang
        >>> ...
        >>> lang = i18nLang(lang_folder='lang')
        >>> welcome = lang.getLanguage('welcome')
        >>> welcome
        >>> 'Hello, Welcome!'

    """

    #: Set the ignored files for the build output.
    __ignore__files: list = [
        '.build.ini', '.i18n',
        'cache', 'config.ini'
    ]

    #: Initialize sidle object
    #: used for getting data and value
    #: on a compiled language pack.
    sidle: Sidle 

    
    def __init__(
        self,
        lang_folder: str,
        lang: Optional[str] = None,
        key: Optional[str] = None
    ):
        """
        Initialize i18nLang folder.

        Parameter:

            `lang_folder` (str):
                The path for the compiled lang folder.

            `default_lang` (Optional[str]):
                The default lang to be used

            `key` (Optional[str]):
                The key for the compiled lang pack.
                This is available @ .build.ini file

        """

        self.lang: str = lang
        self.key:  str = key
        self.version: str = __import__('i18n').lang.__version__
        self.lang_folder = pathlib.Path(lang_folder)

        if not self.key:
            self.key = self.__get_key_from_build_env()
            
        if not self.lang:
            self.lang = self.__get_default_from_build_env()
        else:
            try:
                self.lang = COUNTRIES_TO_LANG[self.lang]
            except KeyError:
                self.lang = self.lang


    def __getitem__(self, key: str):
        """
        Get the lang value of the 
        default language. It can be
        the language or the language code.
        """

        return self.getLanguage(
            key=key
        )

    @logger.catch
    def get(
        self, 
        lang: str,
        key: str,
        error: Optional[bool] = False
    ):

        """
        Get the value of a language either the language
        or the language code from the compiled language
        pack.

        Parameter:
            `lang` (str):
                The language code to be get
            
            `key` (str):
                The key of the language not the key found
                on .build.ini.

        """

        try:
            _lang = COUNTRIES_TO_LANG[lang]
        except KeyError:
            _lang = lang
        
    
        value = self.getLanguage(
            lang=_lang,
            key=key
        )

        if error and not value:
            throw (
                error_cls=KeyError,
                message="Key: '%s' doesnt exist on the language pack" % (key),
                code=ERROR_CODES['not_found']
            )

        return value
    
    @logger.catch
    def getLanguage(
        self,
        key: str
    ) -> str:
        """
        Get the value of a language code from the compiled
        language pack given the lang and the key.

        Parameter:
       
            `key` (str):
                The key of the language not the key found
                on .build.ini.

        """
        files = []

        for file in self.lang_folder.glob('*'):
            if file.name not in self.__ignore__files:
                # TODO: All language file
                files.append(
                    (file.name, str(
                        file.absolute()
                    )
                ))


                if file.name.lower() == self.lang.lower():
                    #: Match the name of the language file
                    #: and the given name
                    self.sidle = Sidle(
                        filename=str(file.absolute()),
                        password=self.key
                    )
                    return self.sidle.__getitem__(
                        key=key
                    )

        
        if self.lang not in files:
            # TODO: No language file registered
            throw(
                error_cls=LanguageError,
                message="Language Code: '%s' not found, You may create a new one and build it" % (self.lang),
                code=ERROR_CODES['not_found']
            )

        return None

    @logger.catch
    def convert_lang(self, lang: str):
        """
        Convert the given language code 
        into the language name. For example:
        `convert_lang('en') -> 'English'`
        """
        
        keys = list(COUNTRIES_TO_LANG.keys())
        values = list(COUNTRIES_TO_LANG.values())
        
        position = values.index(lang)
        return str(keys[position]).capitalize()

    @logger.catch
    def __get_key_from_build_env(self):
        """
        Get the key to access Sidle data on a
        build environment which is `.build.ini`
        File that can be found on a build output 
        folder.
        """

        build_file = self.lang_folder / '.build.ini'

        if not build_file.exists():
            throw(
                error_cls=LanguageError,
                message="Build Config: '.build.ini' not found. You can rebuild if you lost it",
                code=ERROR_CODES['not_found']
            )

        conf = configparser.ConfigParser()
        conf.read(build_file.absolute(), encoding='utf-8')

        return conf.get('i18n', 'key')

    @logger.catch
    def __get_default_from_build_env(self):
        """
        Get the default language registered on a
        build environment which is `.build.ini`
        File that can be found on a build output 
        folder.
        """

        build_file = self.lang_folder / '.build.ini'

        if not build_file.exists():
            throw(
                error_cls=LanguageError,
                message="Build Config: '.build.ini' not found. You can rebuild if you lost it",
                code=ERROR_CODES['not_found']
            )

        conf = configparser.ConfigParser()
        conf.read(build_file.absolute(), encoding='utf-8')
        
        return conf.get('i18n', 'default')


    def __repr__(self):
        """
        The representation of `:class:` i18nLang

        """

        return "%s(version=%s, lang_folder=%s, lang=%s, key=%s)" % (
            type(self).__name__,
            self.version,
            self.lang_folder,
            self.lang, self.key
        ) 

class i18nLangBuilder:
    """

    A i18n Language Builder class that handle different task
    including building and creating i18n environments. This
    class can be used when creating i18n. If you want to use
    the i18n lang compiled project, refer to `:class: i18nLang`.

    Parameters:
    ---
        `lang_folder` (Optional[str]): 
            The name of the environment folder used 
            to store languages
        
        `i18n_env` (Optional[str]):
            Set the name of i18n environment. 
            By default, .i18n is the value.

        `auto_translate` (Optional[bool]):
            Used for generating auto-translated languages
            `generateLang`. By default the value is True

        **options (Keyword Arguments: Dict):
            not yet used. 

    Example:
    ---

        >>> from i18n.lang import i18nLangBuilder
        >>> builder = i18nLangBuilder(...)
        >>> builder.generateLang(lang=['french', 'spanish']) #: For generating
        >>> builder.build() #: For building


    """

    #: The default language code by the system default.
    default_lang = locale.getdefaultlocale()[0]

    #: Set the variable to boolean (True or false)
    #: For defining if the function is reloaded so that
    #: printing banner wont be reloaded as well
    _func_reload: bool = False

    #: Set the build version
    build_version: tuple = (0, 0, 1)

    def __init__(
        self,
        lang_folder: Optional[str] = '_lang',
        i18n_env: Optional[str] = '.i18n',
        auto_translate: Optional[bool] = True,
        **options
    ):

        """
        Initialize `i18nLangBuilder` class.

        Parameters:
    
            `lang_folder` (Optional[str]): 
                The name of the environment folder used 
                to store languages
            
            `i18n_env` (Optional[str]):
                Set the name of i18n environment. 
                By default, .i18n is the value.

            `auto_translate` (Optional[bool]):
                Used for generating auto-translated languages
                `generateLang`. By default the value is True

            **options (Keyword Arguments: Dict):
                not yet used. 
        
        """

        self.lang_folder: str = lang_folder
        self.i18n_env: str = i18n_env
        self._translate: bool = auto_translate
        self.options = options

        if not self.__check_for_lang_folder():
            logger.debug('No i18n environment found, creating new one')
            self.createEnv(self.lang_folder)

    def build(
        self, 
        env_dir: Optional[str] = None, 
        output_dirname: Optional[str] = None,
        clear_cache: Optional[bool] = True
    ):
        """
        Build the _lang folder using Sidle. 
        Make sure the .i18n environment is 
        available. If not then create a new one.

        Parameter:
            `env_dir(Optional[str])`:
                The environment directory. 
                The default value is '_lang'
            
            `output_dirname (Optional[str])`:
                The output directory name after
                building the languages. The default
                value is 'lang'

        """

        if not self._func_reload:
            clear_screen()
            print_banner()
        

        if not env_dir:
            env_dir = pathlib.Path(self.lang_folder)
        else:
            env_dir = pathlib.Path(env_dir)
        
        if not env_dir.exists():
            self.createEnv(
                dir=self.lang_folder
            )

        if not self._func_reload:
            # TODO: Avoid repeating on recursion
            logger.info('Building Folder: %s' % (env_dir))
            print('\n')

        self.__build(
            env_dir=env_dir,
            output_dirname=output_dirname,
            clear_cache=clear_cache
        )

        if not self._func_reload:
            logger.success('Building folder, environment success!')
            logger.success('Enjoy using `i18n.lang`')
            print('')


    def __build(
        self, 
        env_dir: Optional[str] = None, 
        output_dirname: Optional[str] = None,
        clear_cache: Optional[bool] = True
    ):
        """
        A base function for building without
        any progress or any printing output.


        Build the _lang folder using Sidle. 
        Make sure the .i18n environment is 
        available. If not then create a new one.

        Parameter:
            `env_dir(Optional[str])`:
                The environment directory. 
                The default value is '_lang'
            
            `output_dirname (Optional[str])`:
                The output directory name after
                building the languages. The default
                value is 'lang'

        """

        if not self._func_reload:
            # TODO: Avoid repeating on recursion
            logger.info('Reading environment configuration')
        
        i18n_env = (env_dir / self.i18n_env)

        _config = self.read_config(
            path=(i18n_env / 'config.ini')
        )
        
        if not output_dirname:
            _output_dirname = self.__get_config_option(
                config_class=_config,
                section='build',
                option='dirname'
            )
            if _output_dirname:
                output_dirname = _output_dirname
            else:
                output_dirname = 'lang'
        
        key = self.__get_config_option(
            config_class=_config,
            section='build',
            option='key'
        )

        if not key:
            throw(
                error_cls=KeyError,
                message="Build Config: 'key' is not define, kindly create a key option",
                code=ERROR_CODES['not_found']
            )
        
        if not (i18n_env / 'cache').exists():
            # TODO: Create a cache folder for i18n env

            os.mkdir(
                path=(i18n_env / 'cache').absolute()
            )

        for file in env_dir.glob('*'):
            if not file.name == self.i18n_env:
                _parser = self.read_config(
                    path=file.absolute()
                )
                name = self.__get_config_option(
                    config_class=_parser,
                    section='lang',
                    option='name'
                )
                code = self.__get_config_option(
                    config_class=_parser,
                    section='lang',
                    option='code'
                )
                try:
                    metadata = _parser.items('metadata')
               
                except configparser.NoSectionError:
                    throw(
                        error_cls=MetaDataError,
                        message="Config: 'metadata' is not found Please read how to create metadata",
                        code=ERROR_CODES['not_found']
                    )

                # TODO: Encrypt the data using the Sidle with the key given
                try:
                    if not self._func_reload:
                        # TODO: Avoid repeating on recursion
                        logger.info('Building %s -> %s' %(
                                file.name,
                                (i18n_env / 'cache' / file.name).absolute()
                            )
                        )
                    sidle = Sidle(
                        filename=str(
                            (i18n_env / 'cache' / file.name).absolute()
                        ),
                        password=key
                    )

                    if not self._func_reload:
                        # TODO: Avoid repeating on recursion
                        logger.success('Saving %s -> %s' %(
                                file.name,
                                (i18n_env / 'cache' / file.name).absolute()
                            )
                        )
                    if not self._func_reload:
                        # TODO: Avoid repeating on recursion
                        print('')

                except FileNotFoundError:
                    pass

                try:
                    for (k, v) in metadata:
                        
                        sidle['_lang_name'] = name
                        sidle['_lang_code'] = code

                        sidle[k] = v

                        
                except Exception:
                    try:
                        shutil.rmtree(
                            path=str(
                                (i18n_env / 'cache').absolute()
                            )
                        )
                    except FileNotFoundError:
                        pass

                    self._func_reload = True
                    self.build(
                        env_dir=env_dir,
                        output_dirname=output_dirname
                    )


        OUTPUT_DIR = (BASE_DIR / output_dirname)

        if not OUTPUT_DIR.exists():
            if not self._func_reload:
                # TODO: Avoid repeating on recursion
                logger.info('Output Directory: %s doesn\'t exist. Creating a new one')
            
            os.mkdir(
                path = str(
                    OUTPUT_DIR.absolute()
                )
            )
            if not self._func_reload:
                # TODO: Avoid repeating on recursion
                logger.success('Output Directory successfully created')

        cache_folder = (i18n_env / 'cache') 
    
        if not self._func_reload:
            # TODO: Avoid repeating on recursion
            print('')

        for file in cache_folder.glob('*'):
            
            if not self._func_reload:
                # TODO: Avoid repeating on recursion
                logger.info('Moving compiled language: %s to -> %s' % 
                    (
                    file.name, (OUTPUT_DIR / file.name).absolute()
                    )
                )

            shutil.move(
                src=str(file.absolute()),
                dst=str(
                    (OUTPUT_DIR / file.name).absolute()
                )
            )

        if not self._func_reload:
            # TODO: Avoid repeating on recursion
            logger.success('Files moved successfully')

        if not self._func_reload:
            # TODO: Avoid repeating on recursion
            print('')        
        
        if cache_folder.exists() and clear_cache:
            # TODO: Clean cache folder.
            if not self._func_reload:
                # TODO: Avoid repeating on recursion
                logger.info('Cleaning all cache folder..')
                time.sleep(3)

            shutil.rmtree(
                path=str(cache_folder.absolute())
            )

        if not self._func_reload:
            # TODO: Avoid repeating on recursion
            logger.success('Cleaning all cache folder success')


        # TODO: Create build environment.
        if not self._func_reload:
            # TODO: Avoid repeating on recursion

            print('')
            logger.info('Creating build (.build.ini) environment')
            self.__create_build_env(
                path=OUTPUT_DIR.absolute()
            )

            print('')
            print('─'*50)
            logger.success('Building all done!')

    @logger.catch
    def createEnv(
        self, 
        dir: Optional[str] = None, 
        key: Optional[str] = None 
    ):
        """
        Create a i18n environment folder as well as
        create a started `en` lang file.
        """
        if not self._func_reload:
            clear_screen()
            print_banner()
    
        logger.info('Creating a new environment @ %s' % (dir))
        with ui.Spinner('Building configs...', color='cyan') as spin:
            if not dir:
                dir = pathlib.Path(self.lang_folder)
            else:
                dir = pathlib.Path(dir)

            i18n_folder = dir / self.i18n_env

            if not dir.exists():
                os.mkdir(dir.absolute())

            if not (dir / self.i18n_env).exists():
                create_hidden_folder(i18n_folder.absolute())


            if not key:
                key = generate_password()

            _config = configparser.ConfigParser()
            _config['i18n'] = {
                'default': self.default_lang,
                'default_name': self.convert_lang(
                    lang=self.default_lang
                )
            }
            _config['build'] = {
                'dirname': 'lang',
                'key': key
            }

            _config.write(
                open(
                    (i18n_folder / 'config.ini').absolute(),
                    'w',
                    encoding='utf-8'
                )
            )

            _default = configparser.ConfigParser()
            _default['lang'] = {
                'code': self.default_lang,
                'name': self.convert_lang(
                    lang=self.default_lang
                )
            }

            if TRANSLATOR and self._translate:
                _default['metadata'] = {
                    'welcome': TRANSLATOR.translate(
                        text='Hello, Welcome',
                        dest='en_US',
                    ).text
                }

            elif self._translate and not TRANSLATOR:
                # TODO: `googletrans==4.0.0` is not installed
                throw(
                    error_cls=PluginError,
                    message="Plugin: 'googletrans' is not installed, please install googletrans==4.0.0",
                    code=ERROR_CODES['installation_error']
                )


            _default.write(
                open((dir / self.default_lang).absolute(), 'w', encoding='utf-8')
            )

            time.sleep(3)
            spin.stop()
        
        logger.success('Building folder, environment success!')
        logger.success('Check the folder <yellow>.i18n </yellow> to change some configs. Enjoy!`')


    @logger.catch
    def generateLang(self, lang: Any):
        """
        This function is used for generating auto-translated
        language based on `en` lang. You can define the auto-
        translated languages @ `lang` parameter when initializing
        the `:class:` i18Lang.
        """
        if not self._func_reload:
            clear_screen()
            print_banner()
        
        
        _lang = []

        if not self._translate:
            raise LanguageError('You should define `auto_translate` param to True')

        if isinstance(lang, (tuple, list)):
            _lang = lang

        elif isinstance(lang, str):
            _lang.append(lang)
        
        logger.info("Generating Language: '%s'" % (', '.join(_lang)))
        with ui.progressbar(
            total=len(_lang), 
            title='Building Languges ' % (),

        ) as bar:

            for i in _lang:
                i = self.__parse_lang(i)
                self.create_lang(i)
                time.sleep(.005)
                bar()

        hype.print('')
    
    @logger.catch
    def updateLang(self):
        """
        This function update all languages.
        This differ to `generateLang` since `generateLang` 
        create a new one, while this one just update it.
        """

        print_banner()

        path = pathlib.Path(self.lang_folder)
        parser = configparser.ConfigParser()
        new = configparser.ConfigParser()
        _ = []

        for file in path.glob('*'):
            # TODO: Get all files from the lang folder
            if not file.is_dir():
                parser.read(file.absolute(), encoding='utf-8')
                name = self.convert_lang(file.name)
                code = file.name

                try:
                    metadata = parser.items('metadata')
                except configparser.NoSectionError:
                    throw(
                        error_cls=MetaDataError,
                        message="Config: 'metadata' is not found Please read how to create metadata",
                        code=ERROR_CODES['not_found']
                    )
                new['lang'] = {
                    'code': code,
                    'name': name
                }

                for k,v in metadata:
                    if file.name != self.default_lang:
                        try:
                            translated = TRANSLATOR.translate(
                                text=v,
                                dest=file.name,
                                src=self.default_lang
                            )
                        except ValueError:
                            translated = TRANSLATOR.translate(
                                text=v,
                                dest=file.name.replace('_', '-'),
                                src=self.default_lang
                            )

                        _.append((k, translated.text))
                        # print(translated)

                if file.name != self.default_lang:
                    new['metadata'] = {k:v for k,v in _}
                    with open(file.absolute(), 'w', encoding='utf-8') as f:
                        new.write(f)
    
    @logger.catch
    def create_lang(
        self, 
        lang: str, 
        folder: Optional[str] = None
    ):
        """
        Create a lang file with auto translated text. Make sure 
        `.i18n` File exist.
        """


        if not folder:
            folder = self.lang_folder

        path = pathlib.Path(folder)
        i18n_env = path / '.i18n'
        config = i18n_env / 'config.ini'

        if i18n_env.is_file():
            throw(
                error_cls=ConfigError,
                message='Config: %s is a file, please make sure it \n\tis a directory or read how to create .i18n config' % (str(config.name)),
                code=ERROR_CODES['internal_error']
            )

        i18n_config = configparser.ConfigParser()
        i18n_config.read(config.absolute(), encoding='utf-8')
        
        default_lang = self.__get_config_option(
            config_class=i18n_config,
            section='i18n', 
            option='default'
        )

        default_name = self.__get_config_option(
            config_class=i18n_config,
            section='i18n', 
            option='default_nmae'
        )

        default_config = configparser.ConfigParser()
        default_config.read((path / default_lang).absolute())

        lang_config = configparser.ConfigParser()

        metadata = default_config.items('metadata')
        country = self.convert_lang(lang)
        lang_config['lang'] = {'name': country, 'code': lang}
        
        _ = []

    
        for k,v in metadata:

            # TODO: Translate all value from the metadata
            #: For example:
            #:...
            #: [metadata]
            #: greet = Hello
            #:...

            #: Translated
            #: [metadata]
            #: greet = Hola
            try:
                translated = TRANSLATOR.translate(
                    text=v,
                    dest=lang,
                    src=default_lang
                )
            except ValueError:
                translated = TRANSLATOR.translate(
                    text=v,
                    dest=lang.replace('_', '-'),
                    src=default_lang
                )

            _.append((k, translated.text))
            

        lang_config['metadata'] = {k:v for k,v in _}
        lang_config.write(
            open((path / lang).absolute(), 'w', encoding='utf-8')
        )

    @logger.catch
    def read_config(
        self, 
        path: Optional[pathlib.Path],
        allow_no_value: Optional[bool] = False,
        comment_prefixes: Sequence[str] = ('#', ';', '#;')
    ) -> configparser.ConfigParser:

        """
        Read the given config and return -> ConfigParser
        class.
        """

        if not path.exists():
            throw(
                error_cls=FileNotFoundError,
                message="Path: '%s' is not found. Please be specific." % (path),
                code=ERROR_CODES['not_found']
            )

        config = configparser.ConfigParser(
            allow_no_value=allow_no_value,
            comment_prefixes=comment_prefixes
        )

        config.read(path.absolute(), encoding='utf-8')
        
        return config

    def convert_lang(self, lang: str):
        """
        Convert the given language code 
        into the language name. For example:
        `convert_lang('en') -> 'English'`
        """
        
        keys = list(COUNTRIES_TO_LANG.keys())
        values = list(COUNTRIES_TO_LANG.values())
        
        position = values.index(lang)
        return str(keys[position]).capitalize()


    def __check_for_lang_folder(self):
        """
        Check if the lang folder exist or not
        """
        if pathlib.Path(self.lang_folder).exists():
            return True
        else:
            return False
        
    def __parse_lang(self, lang: str):
        """
        Parse the given string to a language code.
        If the value given is country, then it will
        return the language code for the country
        """

        lang = lang.lower()
        
        if len(lang) <= 2:
            # TODO: Language code
            return lang
        
        if lang not in COUNTRIES_TO_LANG.keys():
            # TODO: Check if the given lang exist on the
            #: Dictionary, if not then throw `languageError`

            throw(
                error_cls=LanguageError, 
                message='Language: %s not found, Please be more specific' % (lang.capitalize()),
                code=ERROR_CODES['not_found']
            )
            
        code = COUNTRIES_TO_LANG[lang]
        return code


    def __get_config_option(
        self, 
        config_class: configparser.ConfigParser, 
        section: str, 
        option: str
    ):

        """
        Get the option from the given `ConfigParser` class 
        """

        try:
            data = config_class.get(section, option)
        except Exception:
            data = None

        return data

    def __create_build_env(
        self, 
        path: pathlib.Path
    ):
        """
        Create a build env after building. 
        This will produce .build.ini file
        """

        i18n_build_env = path / '.build.ini'
        i18n_cache = (pathlib.Path(self.lang_folder) / self.i18n_env / 'config.ini')

        i18n_config = self.read_config(
            path=(i18n_cache).absolute()
        )

        
        default_lang = self.__get_config_option(
            config_class=i18n_config,
            section='i18n',
            option='default'
        )

        key = self.__get_config_option(
            config_class=i18n_config,
            section='build',
            option='key'
        )


        _config = configparser.ConfigParser(
            allow_no_value=True,
            comment_prefixes=('#', ';', '#;')
        )
        
        _config['i18n'] = {
            
            #: Add some comments.
            '': None,
            '#; Build Configuration for i18n.lang.': None,
            '#; Please take note that if you want to': None,
            '#; Delete this, make sure to add the key.': None,
            '#; Keys are important. If you accidentally removed,': None,
            '#; Kindly rebuild the i18n language pack.': None,
            '#; Build version: %s' % (''.join(str(self.build_version))): None,
            '': None,
            
            #: Add the key and default lang
            'default': default_lang,
            'key': key
        }

        if i18n_build_env.exists():
            # TODO: If the .build.ini exist, remove all attribute
            hide_file(str(i18n_build_env.absolute()), 0x80)
        
        _config.write(
            open(i18n_build_env.absolute(), 'w', encoding='utf-8')
        )
        

        # TODO: After writting, set the file attribute to
        #: hidden

        hide_file(str(i18n_build_env.absolute()))
    
    def __repr__(self):
        """
        Representation for i18nLangBuilder that display
        its version and some attributes
        """
        return "%s(build_version=%s, i18n_env=%s, lang_folder=%s)" % (
            type(self).__name__, self.build_version, 
            self.i18n_env, self.lang_folder
        )



#               Copyright (c) 2021 Zenqi.

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from setuptools import setup
from setuptools import find_packages
from i18n.lang import __version__

BASE_URL = 'https://github.com/znqi/i18n.lang'

def get_long_description():

    with open("README.md", encoding="utf-8") as f:
        readme = f.read()

    return readme

def get_requirements():
    
    with open('requirements.txt', 'r') as f:
        requirements = f.read().split('\n')
    
    return requirements

extras_require = {
    'translator': 'googletrans==4.0.0rc1'
}

setup(
    
    name="i18n.lang", 
    description="A module used for building and translating data to create application internationally",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    project_urls={
        'Source': BASE_URL,
    },
    author = 'Zenqi',
    license = 'MIT',
    version = __version__,
    install_requires=[
        "colorama>=0.3.4 ; sys_platform=='win32'",
        "aiocontextvars>=0.2.0 ; python_version<'3.7'",
        "win32-setctime>=1.0.0 ; sys_platform=='win32'",
        "pycryptodome>=3.10.1",
        "hypecli>=0.0.6"
    ],
    packages = [p for p in find_packages() if 'test' not in p],
    extras_require = extras_require
)
